#include "StdAfx.h"

#include "../objects/object_dynamic.h"
#include "../objects/object_bullet.h"
#include "../objects/object_item.h"
#include "../objects/object_player.h"
#include "../objects/object_enemy.h"
#include "../objects/object_ray.h"

#include "phys_collisionsolver.h"

#include "sap/OPC_ArraySAP.h"
#include "../object_manager.h"

#include "../editor.h"

//////////////////////////////////////////////////////////////////////////

extern Opcode::ArraySAP *asap;

#ifdef DEBUG_DRAW_COLLISION_PAIRS
extern list<Vector2> collPairs;
#endif // DEBUG_DRAW_COLLISION_PAIRS

extern bool isNewGameCreated;

#ifdef MAP_EDITOR
extern UINT editorAreaSelectorObject;
extern bool editorPhysicProcessed;
extern bool editorAreaSelectNeeded;
#endif // MAP_EDITOR

//////////////////////////////////////////////////////////////////////////

void SolveDynamicVSStatic(ObjDynamic* objD, ObjPhysic* objS);
void SolveBullet(ObjBullet* bul, ObjPhysic* obj);
void SolveEnvironment(ObjEnvironment* env, ObjDynamic* obj);
void SolveDelayedEnvCollision();

bool SolveRay(ObjRay* ray, ObjPhysic* obj);

void ResolveCollisionMTD(ObjDynamic* dyn, const CAABB &bb);
bool ResolveCollision( int sX, int sY, ObjDynamic* dyn, const CAABB &bb );

void ProcessRayCollision();

//////////////////////////////////////////////////////////////////////////

//    , , ,     .
//   ,  - .
list<GameObject*> delayed_env_coll;

#ifdef DEBUG_PRINT
UINT dbgPairsCount = 0;
#endif // DEBUG_PRINT

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

//   callback-      .      .
#if defined(DEBUG_SAP_PAIRS_COORDS)
	#define USE_PAIR_CALLBACK
#endif // defined(DEBUG_SAP_PAIRS_COORDS)

#ifdef USE_PAIR_CALLBACK
void* OnPairCreate(const void* object0, const void* object1, void* user_data)
{
#ifdef DEBUG_SAP_PAIRS_COORDS
	const ObjPhysic* o0 = static_cast<const ObjPhysic*>(object0);
	const ObjPhysic* o1 = static_cast<const ObjPhysic*>(object1);

	sLog(DEFAULT_LOG_NAME, logLevelInfo, "Added pair %d (%.2f, %.2f - %.2f, %.2f) - %d (%.2f, %.2f - %.2f, %.2f)",
		o0->id, o0->aabb.Left(), o0->aabb.Top(), o0->aabb.Right(), o0->aabb.Bottom(), o1->id, o1->aabb.Left(), o1->aabb.Top(), o1->aabb.Right(), o1->aabb.Bottom());
	return NULL;
#endif // DEBUG_SAP_PAIRS_COORDS
}

void  OnPairRemove(const void* object0, const void* object1, void* user_data, void* pair_user_data)
{
#ifdef DEBUG_SAP_PAIRS_COORDS
	const ObjPhysic* o0 = static_cast<const ObjPhysic*>(object0);
	const ObjPhysic* o1 = static_cast<const ObjPhysic*>(object1);

	sLog(DEFAULT_LOG_NAME, logLevelInfo, "Removed pair %d (%.2f, %.2f - %.2f, %.2f) - %d (%.2f, %.2f - %.2f, %.2f)",
		o0->id, o0->aabb.Left(), o0->aabb.Top(), o0->aabb.Right(), o0->aabb.Bottom(), o1->id, o1->aabb.Left(), o1->aabb.Top(), o1->aabb.Right(), o1->aabb.Bottom());
#endif // DEBUG_SAP_PAIRS_COORDS
}
#endif // USE_PAIR_CALLBACK


////////////////////////////////////////////////////////////////////////// 
////////////////////////////////////////////////////////////////////////// 


//      ,       SAP   ,  
//   ProcessCollisions.  -  asap->DumpPairs  asap->UpdateObject,
//       asap->mPairs.mActivePairs    ,
//    ,   ProcessCollisions     .
//    ,  ""  SAP  ,     
// DumpPairs  UpdateObject. ,   ProcessCollisions  "" SAP  
//   .

bool sapStateLocked = false;
bool sapStateUpdatePending = false;
bool sapObjUpdatePending = false;
class UpdateObjInfo
{
public:
	UINT handle;
	ASAP_AABB aabb;
	
	UpdateObjInfo(UINT h, const ASAP_AABB& a) 
		: handle(h), aabb(a)
	{ }


};

list<UpdateObjInfo> updateObjList;

//     DumpPairs(NULL, NULL, NULL, NULL)
void UpdateSAPState()
{
	if (sapStateLocked)
	{
		sapStateUpdatePending = true;
		return;
	}

#ifdef USE_PAIR_CALLBACK
	asap->DumpPairs(OnPairCreate, OnPairRemove, NULL, NULL);
#else
	asap->DumpPairs(NULL, NULL, NULL, NULL);
#endif // USE_PAIR_CALLBACK

	sapStateUpdatePending = false;
}

//     UpdateObject
void UpdateSAPObject(UINT sap_handle, const ASAP_AABB& aabb)
{
	if (sapStateLocked)
	{
		sapObjUpdatePending = true;
		updateObjList.push_back(UpdateObjInfo(sap_handle, aabb));
	}
	else
	{
		asap->UpdateObject(sap_handle, aabb);
	}

}

void BatchSAPObjUpdate()
{
	if (!sapObjUpdatePending)
		return;
	ASSERT(!sapStateLocked);
	ASSERT(!updateObjList.empty());
	for (list<UpdateObjInfo>::iterator it = updateObjList.begin(); it != updateObjList.end(); it++)
	{
		UpdateObjInfo& uoi = *it;
		asap->UpdateObject(uoi.handle, uoi.aabb);
	}
	updateObjList.clear();
	sapObjUpdatePending = false;
}




//////////////////////////////////////////////////////////////////////////

//  
void ProcessCollisions()
{
	Opcode::ASAP_Pair* pairs;

#ifdef DEBUG_SAP
	sLog(DEFAULT_LOG_NAME, logLevelInfo, "Beginning ProcessCollisions");
#endif // DEBUG_SAP

#ifdef DEBUG_DRAW_COLLISION_PAIRS
	collPairs.clear();
#endif // DEBUG_DRAW_COLLISION_PAIRS

#ifdef DEBUG_PRINT
	dbgPairsCount = 0;
#endif // DEBUG_PRINT

	ASSERT(!sapStateLocked);	//   ,   -    ProcessCollisions
	sapStateLocked = true;

#ifdef USE_PAIR_CALLBACK
	UINT pairsCount = asap->DumpPairs(OnPairCreate, OnPairRemove, NULL, &pairs);
#else
	UINT pairsCount = asap->DumpPairs(NULL, NULL, NULL, &pairs);
#endif // USE_PAIR_CALLBACK
	
#ifdef DEBUG_PRINT
	dbgPairsCount = pairsCount;
#endif // DEBUG_PRINT
#ifdef DEBUG_SAP
	sLog(DEFAULT_LOG_NAME, logLevelInfo, "pairs count = %d", pairsCount);
#endif // DEBUG_SAP

	if (pairsCount)
	{
		ObjPhysic** odArr = new ObjPhysic*[pairsCount * 2];
		for(UINT i = 0; i < pairsCount; i++)
		{
			odArr[i*2 + 0] = (ObjPhysic*)pairs[i].object0;
			odArr[i*2 + 1] = (ObjPhysic*)pairs[i].object1;
			//    
			SolveCollision(odArr[i*2 + 0], odArr[i*2 + 1]);
			if (isNewGameCreated)
			{
				DELETEARRAY(odArr);
				delayed_env_coll.clear();
				ASSERT(sapStateLocked);
				sapStateLocked = false;
				updateObjList.clear();
				return;
			}
		}
		for(UINT i = 0, j = 0; i < pairsCount * 2; i+=2)
		{
			j = i+1;
			//     ,     
			asap->UpdateObject(odArr[i]->sap_handle, odArr[i]->aabb.GetASAP_AABB() );
			asap->UpdateObject(odArr[j]->sap_handle, odArr[j]->aabb.GetASAP_AABB() );
		}
		DELETEARRAY(odArr);
	}

	SolveDelayedEnvCollision();
	ProcessRayCollision();

	ASSERT(sapStateLocked);
	sapStateLocked = false;
	if (sapStateUpdatePending) UpdateSAPState();
	BatchSAPObjUpdate();

#ifdef MAP_EDITOR
	if (editorAreaSelectNeeded) editorPhysicProcessed = true;
#endif // MAP_EDITOR

#ifdef DEBUG_SAP
	sLog(DEFAULT_LOG_NAME, logLevelInfo, "Ending ProcessCollisions");
#endif // DEBUG_SAP

}

//////////////////////////////////////////////////////////////////////////

//    ,    
ObjPhysic* arRayObjIntersections[maxRayIntersectionsCount];
//           
float arRayObjIntersectionsDistances[maxRayIntersectionsCount];

//       ,    , 
//    .
UINT GetRayIntersections(bool dir, scalar posFirstCoord, scalar limit, 
			  AxisIndex fa, AxisIndex sa, ObjRay* ray) 
{
	CRay& r = ray->ray;

	const Opcode::ASAP_Box* boxes = asap->GetBoxes();
	const Opcode::ASAP_EndPoint* firstAxis	= asap->GetEndPoints(fa);
	const Opcode::ASAP_EndPoint* secondAxis = asap->GetEndPoints(sa);

	const Opcode::ASAP_EndPoint* current = firstAxis + ( dir ? 0 : (asap->GetNbUsedBoxes()*2 + 1) );
	assert(current->mValue == dir ? MIN_FLOAT : MAX_FLOAT);

	const Opcode::ASAP_Box* object = NULL;
	const Opcode::ASAP_EndPoint* anotherEP;				//   ,    firstAxis
	const Opcode::ASAP_EndPoint *minEP_SA, *maxEP_SA;	//    secondAxis
	scalar leftVal, rightVal, topVal, bottomVal;

	//  ,  , ,  .        
	//     ,  .
	//memset(arRayObjIntersections, 0, sizeof(arRayObjIntersections));
	//memset(arRayObjIntersectionsDistances, 0, sizeof(arRayObjIntersectionsDistances));
	UINT intersCount = 0;
	Vector2 intersPoint;

	//       firstAxis,        
	if (dir)
	{
		//  /

		//  ,  
		// TODO:    ,   
		while((++current)->mValue < posFirstCoord);

		--current;

		//  
		while( (++current)->mValue < limit && intersCount < maxRayIntersectionsCount )
		{
			object = boxes + current->GetOwner();

			if (current->IsMax())
			{
				anotherEP = firstAxis + object->mMin[fa];
				if (anotherEP->mValue <= posFirstCoord)
				{
					//    ,   
					minEP_SA = secondAxis + object->mMin[sa]; 
					maxEP_SA = secondAxis + object->mMax[sa];

					if (fa == X_)
					{
						leftVal		= anotherEP->mValue;
						rightVal	= current->mValue;
						topVal		= minEP_SA->mValue;
						bottomVal	= maxEP_SA->mValue;
					}
					else
					{
						topVal		= anotherEP->mValue;
						bottomVal	= current->mValue;
						leftVal		= minEP_SA->mValue;
						rightVal	= maxEP_SA->mValue;
					}


					if (r.GetIntersectionPoint(leftVal, topVal, rightVal, bottomVal, intersPoint) && 
						ray->CheckIntersection((ObjPhysic*)object->mObject))
					{
						arRayObjIntersections[intersCount] = (ObjPhysic*)object->mObject;
						intersPoint -= r.p;
						arRayObjIntersectionsDistances[intersCount++] = intersPoint.Length();
					}
				}
			}
			else
			{
				//  / , 
				anotherEP	= firstAxis + object->mMax[fa];
				minEP_SA	= secondAxis + object->mMin[sa]; 
				maxEP_SA	= secondAxis + object->mMax[sa];

				if (fa == X_)
				{
					leftVal		= current->mValue;
					rightVal	= anotherEP->mValue;
					topVal		= minEP_SA->mValue;
					bottomVal	= maxEP_SA->mValue;
				}
				else
				{
					topVal		= current->mValue;
					bottomVal	= anotherEP->mValue;
					leftVal		= minEP_SA->mValue;
					rightVal	= maxEP_SA->mValue;
				}


				if (r.GetIntersectionPoint(leftVal, topVal, rightVal, bottomVal, intersPoint) && 
					ray->CheckIntersection((ObjPhysic*)object->mObject))
				{
					arRayObjIntersections[intersCount] = (ObjPhysic*)object->mObject;
					intersPoint -= r.p;
					arRayObjIntersectionsDistances[intersCount++] = intersPoint.Length();
				}
			}
		}
	}
	else
	{
		//  /

		//  ,  
		// TODO:    ,   
		while((--current)->mValue > posFirstCoord);

		++current;

		//  
		while( (--current)->mValue > limit && intersCount < maxRayIntersectionsCount)
		{
			object = boxes + current->GetOwner();

			if (current->IsMax())
			{
				//   , 
				anotherEP	= firstAxis + object->mMin[fa];
				minEP_SA	= secondAxis + object->mMin[sa]; 
				maxEP_SA	= secondAxis + object->mMax[sa];

				if (fa == X_)
				{
					leftVal		= anotherEP->mValue;
					rightVal 	= current->mValue;
					topVal		= minEP_SA->mValue;
					bottomVal	= maxEP_SA->mValue;
				}
				else
				{
					topVal 		= anotherEP->mValue;
					bottomVal 	= current->mValue;
					leftVal		= minEP_SA->mValue;
					rightVal 	= maxEP_SA->mValue;
				}


				if (r.GetIntersectionPoint(leftVal, topVal, rightVal, bottomVal, intersPoint) && 
					ray->CheckIntersection((ObjPhysic*)object->mObject))
				{
					arRayObjIntersections[intersCount] = (ObjPhysic*)object->mObject;
					intersPoint -= r.p;
					arRayObjIntersectionsDistances[intersCount++] = intersPoint.Length();
				}
			}
			else
			{
				anotherEP = firstAxis + object->mMax[fa];
				if (anotherEP->mValue >= posFirstCoord)
				{
					//    ,   
					minEP_SA = secondAxis + object->mMin[sa]; 
					maxEP_SA = secondAxis + object->mMax[sa];

					if (fa == X_)
					{
						leftVal		= current->mValue;
						rightVal	= anotherEP->mValue;
						topVal		= minEP_SA->mValue;
						bottomVal	= maxEP_SA->mValue;
					}
					else
					{
						topVal		= current->mValue;
						bottomVal	= anotherEP->mValue;
						leftVal		= minEP_SA->mValue;
						rightVal	= maxEP_SA->mValue;
					}


					if (r.GetIntersectionPoint(leftVal, topVal, rightVal, bottomVal, intersPoint) && 
						ray->CheckIntersection((ObjPhysic*)object->mObject))
					{
						arRayObjIntersections[intersCount] = (ObjPhysic*)object->mObject;
						intersPoint -= r.p;
						arRayObjIntersectionsDistances[intersCount++] = intersPoint.Length();
					}
				}
			}
		}
	}

	return intersCount;
}

//             SAP
void FindRayAABBintersection(ObjRay* ray)
{
	// TODO:     , 
	ray->debug_color = RGBAf(1.0f, 1.0f, 1.0f, 1.0f);

	// ,      .
	if (ray->activity == oatDying)
		return;

	if(asap->GetNbUsedBoxes() == 0)
		return;

	CRay& r = ray->ray;
	bool dir = r.dir;
	bool isVert = r.IsVertical();
	//bool isHor = r.IsHorizontal();

	assert(asap);	

	//    (,    )  
	//        Y
	AxisIndex fa = isVert ? Y_ : X_;
	AxisIndex sa = isVert ? X_ : Y_;

	scalar posFirstCoord = isVert ? r.p.y : r.p.x;		//     

	scalar limit;		//     firstAxis,     

	if (ray->searchDistance == 0)
	{
		//     -    SAP
		limit = dir ? (MAX_FLOAT - 100) : (MIN_FLOAT + 100);
	}
	else
	{
		//    
		if (isVert)
		{
			//   ,  
			if (dir)
				limit = posFirstCoord + ray->searchDistance;
			else
				limit = posFirstCoord - ray->searchDistance;
		}
		else
		{
			// TODO:  ray->searchDistance / sqrt(r.k*r.k + 1) -   
			//      k -     
			if (dir)
				limit = posFirstCoord + ray->searchDistance / sqrt(r.k*r.k + 1);
			else
				limit = posFirstCoord - ray->searchDistance / sqrt(r.k*r.k + 1);
		}
	}
	
	// TODO:         
	//     (   rev. 1198). 

	UINT intersCount = GetRayIntersections(dir, posFirstCoord, limit, fa, sa, ray);

	if (intersCount > 0)
	{
		//  ,     ,     
		//  ,        
		IceCore::RadixSort r;
		IceCore::RadixSort* RS = &r;
		const UINT* sorted = RS->Sort(arRayObjIntersectionsDistances, intersCount).GetRanks();

		const UINT* const lastSorted = &sorted[intersCount];
		UINT index;
		//  
		while(sorted < lastSorted)
		{
			index =  *sorted++;

			if (SolveRay(ray, arRayObjIntersections[index]))
				break;
		}
	}
}


//        
void ProcessRayCollision()
{
	extern LSQ_HandleT RayTree;
	LSQ_IteratorT rayIter = LSQ_GetFrontElement(RayTree);
	if (rayIter != LSQ_HandleInvalid)
	{
		ObjRay* ray = NULL;
		for (; !LSQ_IsIteratorPastRear(rayIter); LSQ_AdvanceOneElement(rayIter))
		{
			ray = ((ObjRay*)LSQ_DereferenceIterator(rayIter));
			assert(ray);
			FindRayAABBintersection(ray);
		}
		LSQ_DestroyIterator(rayIter);
	}
}

//  true,        .
bool SolveRay(ObjRay* ray, ObjPhysic* obj)
{
	// TODO:      
	ray->debug_color = RGBAf(1.0f, 0.2f, 0.2f, 1.0f);

	if (ray->Hit(obj))
	{
		ray->debug_color = RGBAf(0.2f, 1.0f, 0.2f, 1.0f);
	}

	return ray->activity == oatDying;
}

//////////////////////////////////////////////////////////////////////////


//,     .
bool SolveVSSlope(ObjPhysic* obj, ObjPhysic* slope, Vector2 shift )
{
	if (!slope || !slope->slopeType) return false;

	if (!Collide(obj->aabb, slope->aabb))
		return false;

	if ( obj->aabb.Bottom() > slope->aabb.Bottom() ) return false;

	//if (obj->aabb.p.x+shift.x < slope->aabb.Left() || obj->aabb.p.x + shift.x > slope->aabb.Right() ) return false;
	//  /|
	// /_|
	if (slope->slopeType == 1)
	{
		if (obj->aabb.Bottom()+shift.y > slope->aabb.Bottom() - 2*slope->aabb.H*((obj->aabb.p.x+shift.x-slope->aabb.Left())/(2*slope->aabb.W)) &&
			obj->aabb.p.x+shift.x > slope->aabb.Left()+2*slope->aabb.W/(obj->aabb.Bottom()+shift.y-slope->aabb.Bottom()))
			return true;
		return false;
	}
	// |\ /
	// |_\/
	if (slope->slopeType == 2)
	{
		if (obj->aabb.Bottom()+shift.y > slope->aabb.Bottom() - (2*slope->aabb.H)*((slope->aabb.Right()-obj->aabb.p.x+shift.x)/(2*slope->aabb.W)) &&
			obj->aabb.p.x+shift.x < slope->aabb.Right()-2*slope->aabb.W/(obj->aabb.Bottom()+shift.y-slope->aabb.Bottom()))
			return true;
		return false;
	}
	return false;
}

float GetSlopeTop(ObjPhysic* slope, float x )
{
	if (!slope) return false;
	//  /|
	// /_|
	if (slope->slopeType == 1)
		return slope->aabb.Bottom()-2*slope->aabb.H*((x-slope->aabb.Left())/(2*slope->aabb.W));
	// |\ /
	// |_\/
	if (slope->slopeType == 2)
		return slope->aabb.Bottom()-2*slope->aabb.H*((slope->aabb.Right()-x)/(2*slope->aabb.W));
	return false;
}

float GetSlopeX(ObjPhysic* slope, float y )
{
	if (!slope) return false;
	return slope->aabb.Left()+(y-slope->aabb.Bottom())/(2*slope->aabb.H);
}

bool TouchCondition( ObjDynamic* toucher, GameObject* obj )
{
	if ( obj->IsSleep() || toucher->IsSleep() ) return false;
	switch (obj->touch_detection)
	{
		case tdtFromEverywhere:
			{
				if (toucher->aabb.p.x - toucher->vel.x > obj->aabb.Right() || toucher->aabb.p.x - toucher->vel.x < obj->aabb.Left())
					return true;
				if (toucher->aabb.p.y - toucher->vel.y > obj->aabb.Bottom() || toucher->aabb.p.y - toucher->vel.y < obj->aabb.Top())
					return true;
				return false;
			}
		case tdtFromTop:
			{
				if (toucher->vel.y > 0 && toucher->aabb.p.y - toucher->vel.y < obj->aabb.Top() )
					return true;
				return false;
			}
		case tdtFromBottom:
			{
				if (toucher->vel.y < 0 && toucher->aabb.p.y - toucher->vel.y > obj->aabb.Bottom() )
					return true;
				return false;
			}
		case tdtFromSides:
			{
				if (toucher->aabb.p.x - toucher->vel.x > obj->aabb.Right() || toucher->aabb.p.x - toucher->vel.x < obj->aabb.Left())
					return true;
				return false;
			}
		case tdtTopAndSides:
			{
				if (toucher->aabb.p.y - toucher->vel.y > obj->aabb.Bottom())
					return false;
				return true;
			}
		default:
			return true;
	}
}

bool SolidTo( ObjPhysic* obj, ObjectType type )
{
	switch (type)
	{
		case objPlayer:
			return (obj->solid_to & 1) != 0;
		case objBullet:
			return (obj->solid_to & 2) != 0;
		case objEffect:
			return (obj->solid_to & 4) != 0;
		case objSprite:
			return (obj->solid_to & 8) != 0;
		case objItem:
			return (obj->solid_to & 16) != 0;
		case objEnemy:
			return (obj->solid_to & 32) != 0;
		default:
			return true;
	}
}

void SolveCollision(ObjPhysic* obj1, ObjPhysic* obj2)
{
	assert(obj1);
	assert(obj2);

#ifdef MAP_EDITOR
	if (obj1->id == editorAreaSelectorObject)
	{
		AddObjectInArea(obj2->id);
	}
	else if (obj2->id == editorAreaSelectorObject)
	{
		AddObjectInArea(obj1->id);
	}

	return;
#endif

	if ( !SolidTo(obj1, obj2->type) || !SolidTo(obj2, obj1->type) )
	{
		return;
	}

#ifdef DEBUG_DRAW_COLLISION_PAIRS
	collPairs.push_back(obj1->aabb.p);
	collPairs.push_back(obj2->aabb.p);
#endif // DEBUG_DRAW_COLLISION_PAIRS

	if ( (obj1->parentConnection && obj1->parentConnection->getParent() == obj2)
		|| (obj2->parentConnection && obj2->parentConnection->getParent() == obj1) )
	{
		asap->RemovePair(obj1, obj2, obj1->id, obj2->id);
	}
	else if (obj1->IsDynamic() && obj2->IsDynamic())
	{ 
		if (((ObjDynamic*)obj1)->IsBullet() )
		{
			SolveBullet((ObjBullet*)obj1, obj2);
			return;
		}

		if (((ObjDynamic*)obj2)->IsBullet() )
		{
			SolveBullet((ObjBullet*)obj2, obj1);
			return;
		}

		if ( !obj1->IsSolid() && obj2->IsSolid() ) //   ,    
			SolveDynamicVSStatic( (ObjDynamic*)obj1, obj2 );
		else if ( obj1->IsSolid() && !obj2->IsSolid() )
			SolveDynamicVSStatic( (ObjDynamic*)obj2, obj1 );

		if ( obj1->type == objPlayer && obj2->IsTouchable() && TouchCondition((ObjDynamic*)obj1, obj2) ) //    . .
		{
			switch ( obj2->type )
			{
				case objItem:
					((ObjItem*)obj2)->Touch( (ObjDynamic*)obj1 );
					return;
				case objEnemy:
					((ObjEnemy*)obj2)->Touch( (ObjDynamic*)obj1 );
					return;
				default:
					obj2->Touch( (ObjDynamic*)obj1 );
					return;
			}
		}
		if ( obj2->type == objPlayer && obj1->IsTouchable() && TouchCondition((ObjDynamic*)obj2, obj1) ) //    . .
		{
			switch ( obj1->type )
			{
				case objItem:
					((ObjItem*)obj1)->Touch( (ObjDynamic*)obj2 );
					return;
				case objEnemy:
					((ObjEnemy*)obj1)->Touch( (ObjDynamic*)obj2 );
					return;
				default:
					obj1->Touch( (ObjDynamic*)obj2 );
					return;
			}
		}

	}
	else if (obj1->IsDynamic() && !obj2->IsDynamic())
	{
		CAABB old_aabb = obj1->aabb;
		if (((ObjDynamic*)obj1)->IsBullet() )
			SolveBullet((ObjBullet*)obj1, obj2);
		else 
			SolveDynamicVSStatic((ObjDynamic*)obj1, obj2);
		
		if (obj1->childrenConnection && obj1->aabb != old_aabb)
			obj1->childrenConnection->Event( ObjectEventInfo( eventCollisionPushed, obj1->aabb.p-old_aabb.p));
	}
	else if (!obj1->IsDynamic() && obj2->IsDynamic())
	{
		CAABB old_aabb = obj2->aabb;
		if (((ObjDynamic*)obj1)->IsBullet() )
			SolveBullet((ObjBullet*)obj2, obj1);
		else 
			SolveDynamicVSStatic((ObjDynamic*)obj2, obj1);
		
		if (obj2->childrenConnection && obj1->aabb != old_aabb)
			obj2->childrenConnection->Event( ObjectEventInfo( eventCollisionPushed, obj2->aabb.p-old_aabb.p));
	}
	else
	{
		asap->RemovePair(obj1, obj2, obj1->id, obj2->id);
	}
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

//      .
void SolveDynamicVSStatic(ObjDynamic* objD, ObjPhysic* objS)
{
	//           .
	if ( objS->IsOneSide() && objD->gravity == Vector2::Blank() )
		return;

	//   
	if ( objS->type == objEnvironment )
	{
		SolveEnvironment((ObjEnvironment*)objS, objD);
		return;
	}

	//   
	if ( objS->slopeType > 0 )
	{
		if ( objD->type == objBullet ) return; //    
		//											    V-   -
		if (SolveVSSlope( objD, objS, Vector2(0,0)) /*&& (objD->vel.y == 0 || !SolveVSSlope( objD, objS, -1*objD->vel ) )*/)
		{
			float slopeY = GetSlopeTop(objS, objD->aabb.p.x) - objD->aabb.H;
			if (objD->aabb.p.y > slopeY) objD->aabb.p.y = slopeY;
			objD->SetOnPlane();	
			objD->vel.y = 0;
			objD->suspected_plane = objS;
		}
		return;
	}
	// 

	if (!objS->IsSolid())
		return;

	if (objD->ghostlike)
		return;

	bool collided = false;
	CAABB tmp(objD->aabb);
	tmp.p -= objD->vel;

	int SegStateX = SegProjIntersect(Vector2(tmp.p.x - tmp.W, tmp.p.x + tmp.W),
		Vector2(objS->aabb.p.x - objS->aabb.W, objS->aabb.p.x + objS->aabb.W));

	int SegStateY = SegProjIntersect(Vector2(tmp.p.y - tmp.H, tmp.p.y + tmp.H),
		Vector2(objS->aabb.p.y - objS->aabb.H, objS->aabb.p.y + objS->aabb.H));

	int oSx = SegProjIntersect(Vector2(objD->old_aabb.p.x - objD->old_aabb.W, objD->old_aabb.p.x + objD->old_aabb.W),
		Vector2(objS->aabb.p.x - objS->aabb.W, objS->aabb.p.x + objS->aabb.W));

	int oSy = SegProjIntersect(Vector2(objD->old_aabb.p.y - objD->old_aabb.H, objD->old_aabb.p.y + objD->old_aabb.H),
		Vector2(objS->aabb.p.y - objS->aabb.H, objS->aabb.p.y + objS->aabb.H));

	//       , ....
	if (SegStateX == SEG_PROJ_STATE_INTERSECT && SegStateY == SEG_PROJ_STATE_INTERSECT && !(objS->IsOneSide()))
	{
		// ...    
		//        ?!+
		if ((oSx != SEG_PROJ_STATE_INTERSECT) && (oSy != SEG_PROJ_STATE_INTERSECT))
			CornerProblemSolve(oSx, oSx, tmp, objS->aabb, objD->vel);

		if (!(oSx == SEG_PROJ_STATE_INTERSECT && oSy == SEG_PROJ_STATE_INTERSECT))
		{
			if (ResolveCollision(oSx, oSy, objD, objS->aabb)) objD->suspected_plane = objS;
		}
		else
			//     ,     
			ResolveCollisionMTD(objD, objS->aabb);
		//      ...
		tmp = CAABB(objD->aabb);
		//oSx = SegStateX;
		//oSy = SegStateY;
		return;
	}

	if (Collide(objD->aabb, objS->aabb))
	{
		collided = true;

		//  ,    ,     .
		if ((SegStateX != SEG_PROJ_STATE_INTERSECT) && (SegStateY != SEG_PROJ_STATE_INTERSECT) && !(objS->IsOneSide()))
			CornerProblemSolve(SegStateX, SegStateY, tmp, objS->aabb, objD->vel);

		if ( (objS->IsOneSide() && objD->IsDropping() && objS->IsForced() && objD->type == objPlayer)
			|| (!objS->IsOneSide() && objD->IsDropping()))
			{
				//  ,     .
				((ObjCharacter*)objD)->movement = omtDropSitting;
			}
		if ( objS->IsOneSide() && !(objD->IsDropping() && objS->IsForced()) )
		{
			if ( ((objD->aabb.p.y)-2*(objD->vel.y)+(objD->aabb.H)-1 <= (objS->aabb.p.y)-(objS->aabb.H)) && (objD->vel.y > 0 && (!objD->IsDropping() || objS->IsForced()) )
				&& !(objD->IsBullet() && objD->bounce+objD->env->bounce_bonus ==0 ))
			{
					BOUNCE_Y(objD);
					objD->aabb.p.y =objS->aabb.p.y - objS->aabb.H - objD->aabb.H - 1;
					objD->SetOnPlane();
					objD->suspected_plane = objS;
			}
		}
		else if (ResolveCollision(SegStateX, SegStateY, objD, objS->aabb)) objD->suspected_plane = objS;
		//      ...
		//tmp = AABB(dyn->aabb);
	}
}

void SolveBullet(ObjBullet* bul, ObjPhysic* obj)
{
	if ( obj->slopeType > 0 )
	{
		if (SolveVSSlope(bul, obj, Vector2(0,0)))
		{
			bul->Hit(obj);
		}
	}
	else bul->Hit(obj);
	//    ,      .   ,   -   .
	if ( bul->activity != oatDying ) SolveDynamicVSStatic( bul, obj );
}

void SolveEnvironment(ObjEnvironment* env, ObjDynamic* obj)
{
	
	if(env == obj->old_env)
	{
		obj->env = env;
		env->OnStay(obj);
	}
	else
	{
		//      
		delayed_env_coll.push_back(env);
		delayed_env_coll.push_back(obj);
	}
	

}

//     .   ,  
//      ,   .
void SolveDelayedEnvCollision()
{
	if (delayed_env_coll.empty())
		return;

	ObjDynamic* obj = NULL;
	ObjEnvironment* env = NULL;
	list<GameObject*>::iterator it = delayed_env_coll.begin();
	
	while(it != delayed_env_coll.end())
	{
		env = (ObjEnvironment*)*it;
		it++;
		obj = (ObjDynamic*)*it;
		it++;

		if (Collide(env->aabb, obj->aabb))
		{
			//      .
			obj->env = env;
			env->OnEnter( obj );
			if ( obj->old_env != default_environment ) 
				obj->old_env->OnLeave( obj );
		}
	}

	delayed_env_coll.clear();
}

//////////////////////////////////////////////////////////////////////////

//    MTD MTD = Minimal Translation Distance
//    void PhysDynamicBody::SatanaResolveCollisionMTD(const ASAP_AABB &bb)
void ResolveCollisionMTD(ObjDynamic* dyn, const CAABB &bb)
{
	scalar deltaX, deltaY;
	SegIntersect(	Vector2(dyn->aabb.p.x - dyn->aabb.W, dyn->aabb.p.x + dyn->aabb.W),
		Vector2(bb.p.x - bb.W, bb.p.x + bb.W), deltaX	);
	SegIntersect(	Vector2(dyn->aabb.p.y - dyn->aabb.H, dyn->aabb.p.y + dyn->aabb.H),
		Vector2(bb.p.y - bb.H, bb.p.y + bb.H), deltaY	);
	if (fabs(deltaX) < fabs(deltaY))
	{
		dyn->aabb.p.x += deltaX;
		dyn->vel.x = 0.0f;
	}
	else
	{
		dyn->aabb.p.y += deltaY;
		dyn->vel.y = 0.0f;
	}
}

//    
bool ResolveCollision( int sX, int sY, ObjDynamic* dyn, const CAABB &bb )
{
	bool mark_as_suspected = false;
	if (sX == SEG_PROJ_STATE_INTERSECT)
	{
		if (sY == SEG_PROJ_STATE_BEFORE)
		{
			BOUNCE_Y(dyn);
			dyn->aabb.p.y = bb.p.y - bb.H - dyn->aabb.H - 1;
			dyn->SetOnPlane();
			mark_as_suspected = true;
		}
		if (sY == SEG_PROJ_STATE_AFTER)
		{
			BOUNCE_Y(dyn);
			dyn->aabb.p.y = bb.p.y + bb.H + dyn->aabb.H + 1;
			dyn->ClearOnPlane();
		}
	}

	if (sY == SEG_PROJ_STATE_INTERSECT)
	{
		if (sX == SEG_PROJ_STATE_BEFORE)
		{
			BOUNCE_X(dyn);
			dyn->aabb.p.x = bb.p.x - bb.W - dyn->aabb.W - 1;
		}
		if (sX == SEG_PROJ_STATE_AFTER)
		{
			BOUNCE_X(dyn);
			dyn->aabb.p.x = bb.p.x + bb.W + dyn->aabb.W + 1;
		}
	}

	if (sY == SEG_PROJ_STATE_BEFORE)
	{
		if (sX == SEG_PROJ_STATE_BEFORE )
		{

			BOUNCE_X(dyn);
			BOUNCE_Y(dyn);
			dyn->aabb.p.x = bb.p.x - bb.W - dyn->aabb.W - 1;
			dyn->aabb.p.y = bb.p.y - bb.H - dyn->aabb.H - 1;
		}
		if (sX == SEG_PROJ_STATE_AFTER )
		{

			BOUNCE_X(dyn);
			BOUNCE_Y(dyn);
			dyn->aabb.p.x = bb.p.x + bb.W + dyn->aabb.W + 1;
			dyn->aabb.p.y = bb.p.y - bb.H - dyn->aabb.H - 1;
		}
	}

	if (sY == SEG_PROJ_STATE_AFTER)
	{
		if (sX == SEG_PROJ_STATE_BEFORE )
		{

			BOUNCE_Y(dyn);
			dyn->aabb.p.y = bb.p.y + bb.H + dyn->aabb.H + 1;
		}
		if (sX == SEG_PROJ_STATE_AFTER )
		{

			BOUNCE_Y(dyn);
			dyn->aabb.p.y = bb.p.y + bb.H + dyn->aabb.H + 1;
		}
	}
	return mark_as_suspected;
}
